@@ -30,7 +30,6 @@
 #include <config.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <ctype.h>
 #include <string.h>
 #include <stdarg.h>
 #include <locale.h>
@@ -1398,7 +1397,12 @@ gnome_font_face_ps_embed_tt (GnomeFontPsObject *pso)
 	nglyphs = pso->face->num_glyphs;
 
 	len = pso->encodedname ? strlen (pso->encodedname) : 0;
-	lower = (len > 3) ? atoi (pso->encodedname + len - 3) : 0;
+
+	if (len > 4)
+		lower = *(pso->encodedname + len - 4) == '_' ? atoi (pso->encodedname + len - 3) : 0;
+	else
+		lower = 0;
+
 	upper = lower + 1;
 
 	k = 1;
