@@ -28,6 +28,7 @@
 #include <config.h>
 #include <string.h>
 #include <gmodule.h>
+#include <errno.h>
 
 #include <libgnomeprint/gnome-print-i18n.h>
 #include <libgnomeprint/gnome-print-private.h>
@@ -41,13 +42,26 @@
 #include <libgnomeprint/gnome-print-svg.h>
 #endif
 
+#include <sys/types.h>
+#include <sys/stat.h>
+
+#include <fcntl.h>
+
 #ifdef HAVE_MMAP
 /* For the buffer stuff, remove when the buffer stuff is moved out here */
-#include <sys/types.h>
 #include <sys/mman.h>
-#include <sys/stat.h>
-#include <fcntl.h>
 #include <unistd.h>
+
+#ifndef PROT_READ
+#define PROT_READ 0x1
+#endif
+
+#if defined(FREEBSD) || defined(__FreeBSD__)
+/* We must keep the file open while pages are mapped.  */
+/* http://www.freebsd.org/cgi/query-pr.cgi?pr=48291 */
+#define HAVE_BROKEN_MMAP
+#endif
+
 #endif /* HAVE_MMAP */
 
 static void gnome_print_context_class_init (GnomePrintContextClass *klass);
@@ -528,29 +542,37 @@ gnome_print_context_new (GnomePrintConfig *config)
 void
 gnome_print_buffer_munmap (GnomePrintBuffer *b)
 {
-#ifndef HAVE_MMAP
-	g_warning ("mmap not supported");
-#else
-	if (b->buf)
-		munmap (b->buf, b->buf_size);
+	if (b->buf) {
+#ifdef HAVE_MMAP
+		if (b->was_mmaped)
+			munmap (b->buf, b->buf_size);
+		else
+#endif
+			g_free (b->buf);
+	}
+
 	b->buf = NULL;
 	b->buf_size = 0;
+
+#ifdef HAVE_BROKEN_MMAP
+	if (b->fd != -1)
+		close (b->fd);
 #endif
 }
 
+#define GP_MMAP_BUFSIZ 4096
+
 gint
 gnome_print_buffer_mmap (GnomePrintBuffer *b,
 			 const guchar *file_name)
 {
-#ifndef HAVE_MMAP
-	g_warning ("Can't mmap file %s", file_name);
-	return GNOME_PRINT_ERROR_UNKNOWN;
-#else
 	struct stat s;
 	gint fh;
 
 	b->buf = NULL;
 	b->buf_size = 0;
+	b->was_mmaped = FALSE;
+	b->fd = -1;
 
 	fh = open (file_name, O_RDONLY);
 	if (fh < 0) {
@@ -559,18 +581,57 @@ gnome_print_buffer_mmap (GnomePrintBuffer *b,
 	}
 	if (fstat (fh, &s) != 0) {
 		g_warning ("Can't stat \"%s\"", file_name);
+		close (fh);
 		return GNOME_PRINT_ERROR_UNKNOWN;
 	}
 	
+#ifdef HAVE_MMAP
 	b->buf = mmap (NULL, s.st_size, PROT_READ, MAP_SHARED, fh, 0);
-	b->buf_size = s.st_size;
-	
-	close (fh);
+#endif
+
+	if ((b->buf == NULL) || (b->buf == (void *) -1)) {
+		g_warning ("Can't mmap file %s - attempting a fallback...", file_name);
+
+		b->buf = g_try_malloc (s.st_size);
+		b->buf_size = s.st_size;
+		if (b->buf != NULL) {
+			ssize_t nread, total_read;
+			
+			nread = total_read = 0;
+			
+			while (total_read < s.st_size) {
+				nread = read (fh, b->buf + total_read,
+					      MIN (GP_MMAP_BUFSIZ, s.st_size - total_read));
+				if (nread == 0) {
+					b->buf_size = total_read;
+					break; /* success */
+				} else if (nread == -1) {
+					if (errno != EINTR) {
+						g_free (b->buf);
+						b->buf = NULL;
+						b->buf_size = 0;
+						break; /* failure */
+				}
+				} else 
+					total_read += nread;
+			}
+		}
+	} else {
+		b->was_mmaped = TRUE;
+		b->buf_size = s.st_size;
+	}
+
+#ifdef HAVE_BROKEN_MMAP
+	if (b->buf != NULL)
+		b->fd = fh;
+	else
+#endif
+		close (fh);
+
 	if ((b->buf == NULL) || (b->buf == (void *) -1)) {
 		g_warning ("Can't mmap file %s", file_name);
 		return GNOME_PRINT_ERROR_UNKNOWN;
 	}
 
 	return GNOME_PRINT_OK;
-#endif /* HAVE_MMAP */
 }
